#!/usr/bin/perl -w

# Copyright 2013, 2014 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Draw distance matrix colours per
#
#     W. Kuhirun, "Simple Procedure for Evaluating the Impedance Matrix of
#     Fractal and Fractile Arrays", Progress In Electromagnetics Research M,
#     volume 14, 2010, pages 61-70.
#
# Points in the terdragon are rotated to have the endpoint horizontal, then
# numbered from Y top downwards and among equal Y with X increasing to the
# right.
#
# Points in the flowsnake are numbered along the curve, no rotation
# necessary;

use 5.010;
use strict;
use Image::Base::GD;
use List::Util 'min','max';
use Math::PlanePath::TerdragonCurve 117; # v.117 for level_to_n_range()
use Math::PlanePath::Flowsnake 117;      # v.117 for level_to_n_range()
use Math::Complex;
use Tie::IxHash;

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyTerdragon ':all';

my $width = 800;
my $height = 800;

# 0 <= $f <= 1, return a colour string "#AABBCC"
sub colour_heat {
  my ($f) = @_;
  return rgb1_to_rgbstr (map { ($f < $_        ? 0
                                : $f < $_+.25  ? 4*($f-$_)
                                : $f < $_+.5   ? 1
                                : $f < $_+.75  ? 4*($_+.75 - $f)
                                : 0)
                             } .375, .125, -.125);
}
sub rgb1_to_rgbstr {
  return sprintf("#%02X%02X%02X",
                 map { max (0, min (0xFF, int (0.5 + 0xFF * $_))) }
                 @_);
}

sub distance {
  my ($x1,$y1, $x2,$y2) = @_;
  my $dx = $x1 - $x2;
  my $dy = $y1 - $y2;
  return sqrt($dx*$dx + $dy*$dy);
}
my @filenames;
CURVE: foreach my $curve ('TerdragonCurve',
                          'Flowsnake',
                         ) {
  my $class = "Math::PlanePath::$curve";
  my $path = $class->new;

  foreach my $level (0 .. 6) {
    print "$curve $level\n";
    my ($n_lo,$n_hi) = $path->level_to_n_range($level);
    my ($ex,$ey) = $path->n_to_xy($n_hi);

    my $rotate = sub {
      my ($x,$y) = @_;

      # ($x + I*$y*sqrt(3)) / ($ex + I*$ey*sqrt(3))
      # = ($x + I*$y*sqrt(3)) * ($ex - I*$ey*sqrt(3))  / enorm
      # = $x*$ex + 3*$y*$ey,  ($x*$ey - $y*$ex)*sqrt(3)

      return ($x*$ex + 3*$y*$ey,
              $x*$ey - $y*$ex);
    };

    if ($curve eq 'Flowsnake') {
      $rotate = sub { @_ };
    }

    my %seen;
    tie %seen, 'Tie::IxHash';
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      $seen{"$x,$y"} = 1;
    }

    my @rotated = map {
      my ($x,$y) = split /,/, $_;
      ($x,$y) = $rotate->($x,$y);
      [$x,$y]
    } keys %seen;

    if ($curve eq 'TerdragonCurve') {
      @rotated = sort { ($a->[1] <=> $b->[1])      # Y descending
                          || ($a->[0] <=> $b->[0]) # X ascending
                        } @rotated;
    }
    if ($curve eq 'Flowsnake') {
      @rotated = reverse @rotated;
    }

    print "  start ",$rotated[0]->[0],", ",$rotated[0]->[1],"\n";
    print "  end   ",$rotated[-1]->[0],", ",$rotated[-1]->[1],"\n";

    my $Np = scalar(@rotated);
    if ($Np > min($width,$height)) {
      print "$curve stop at level=$level Np=$Np\n";
      next CURVE;
    }
    my $scale = max(1, int(min($width,$height)/$Np));

    my $x_origin = ($width - $scale*$Np) / 2;
    my $y_origin = ($height - $scale*$Np) / 2;

    my $image = Image::Base::GD->new (-width => $width, -height => $height);
    $image->rectangle (0,0, $width-1,$height-1, 'black');

    my $max_dist = 0;
    foreach my $i (0 .. $#rotated) {
      foreach my $j (0 .. $#rotated) {
        my $d = distance(@{$rotated[$i]}, @{$rotated[$j]});
        $max_dist = max($max_dist, $d);
      }
    }

    foreach my $i (0 .. $#rotated) {
      foreach my $j (0 .. $#rotated) {
        my $x = $x_origin + $i*$scale;
        my $y = $y_origin + $j*$scale;
        my $d = distance(@{$rotated[$i]}, @{$rotated[$j]});
        my $colour = colour_heat($d/$max_dist);
        $image->rectangle ($x,$y, $x+$scale-1,$y+$scale-1, $colour, 1);
      }
    }

    my $filename = "/tmp/$curve-$level.png";
    $image->save($filename);
    push @filenames, $filename;
  }
}


require IPC::Run;
IPC::Run::run(['xzgv',@filenames],'&');
exit 0;
