/* Copyright 2016, 2017, 2018 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Some terdragon graph stuff using Nauty. */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* GP-DEFINE  read("terdragon-defines.gp"); */

/* ------------------------------------------------------------------------- */
/* Misc Helpers */

/* Change (x,y) to ((x+3y)/2, (x-y)/2) which is a rotation by +120 degrees. */
#define TERDRAGON_XY_ROTATE_PLUS120(x,y)                \
  do {                                                  \
    int TERDRAGON_XY_ROTATE_PLUS120__x = (x);           \
    int TERDRAGON_XY_ROTATE_PLUS120__y = (y);           \
    (x) = (TERDRAGON_XY_ROTATE_PLUS120__x               \
           + 3*TERDRAGON_XY_ROTATE_PLUS120__y) / -2;    \
    (y) = (TERDRAGON_XY_ROTATE_PLUS120__x               \
           - TERDRAGON_XY_ROTATE_PLUS120__y) / 2;       \
  } while (0)

/* Change (x,y) to ((3y-x)/2, (x+y)/2) which is a rotation by -120 degrees. */
#define TERDRAGON_XY_ROTATE_MINUS120(x,y)               \
  do {                                                  \
    int TERDRAGON_XY_ROTATE_PLUS120__x = (x);           \
    int TERDRAGON_XY_ROTATE_PLUS120__y = (y);           \
    (x) = (3*TERDRAGON_XY_ROTATE_PLUS120__y             \
           - TERDRAGON_XY_ROTATE_PLUS120__x) / 2;       \
    (y) = (TERDRAGON_XY_ROTATE_PLUS120__x               \
           + TERDRAGON_XY_ROTATE_PLUS120__y) / -2;      \
  } while (0)


/* ------------------------------------------------------------------------- */
/* Terdragon Sequence Values */

void
terdragon_overflow (void)
{
  fprintf(stderr, "terdragon numeric overflow\n");
  abort();
}

/* terdragon_A(k) returns the area (number of unit triangles) enclosed by
   terdragon k.
   A(k) = if(k==0,0, 2*( 3^(k-1)-2^(k-1) ))
*/
int
terdragon_A (int k)
{
  double d;
  int A;
  if (k==0) return 0;
  d = 2*pow(3.0,k-1) - pow(2,k);
  A = d;
  if (A < d) { terdragon_overflow(); }
  return A;
}

/* terdragon_P(k) returns the number distinct points visited by terdragon curve k.
   2, 4, 8, 18, 44, 114, 308, 858
   P(k) = 3^(k-1) + 2^k + 1
*/
int
terdragon_P (int k)
{
  double d;
  int P;
  if (k==0) return 2;
  d = pow(3,k-1) + (1 << k) + 1;
  P = d;
  if (P < d) { terdragon_overflow(); }
  return P;
}

int
terdragon_TP (int k)
{
  if (k==0) return 3;
  return (int) pow(3,k) + 3*(1<<(k-1)) + 1;
}

/* Return true if the terdragon curve turns right after segment n.
   The first segment is n=0.  */
int
terdragon_next_turn_is_right (int n)
{
  while ((n%3)==2) {
    n = (n-2)/3;
  }
  return (n%3)==1;
}

/* ------------------------------------------------------------------------- */
/* Terdragon Curve Graph */

/* terdragon_graph(g, m,n, k) sets graph g to terdragon curve k.
   n should be the number of vertices terdragon_P(k).
   g should have space for size m,n.

   The current implementation follows the curve by its turn sequence and
   does a linear search for preceding x,y to notice a point revisited.  This
   is not very efficient but is easy and is fast enough since any subsequent
   real work on the graph takes much longer.
 */
int terdragon_graph_INTERNAL_last_v;
static void
terdragon_graph_INTERNAL(graph *g, int m,int n, int start,int end)
{
  struct { int x; int y; } *v_to_xy;
  int i;
  int from = -1;
  int dx = 2;
  int dy = 0;
  int xy_upto = 0;
  int x = 0;
  int y = 0;
  DEBUG(printf("terdragon_graph_INTERNAL n=%d  start=%d end=%d\n", n, start,end));

  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("terdragon_graph()");

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  for (i = 0; i <= end; i++) {
    int to;
    if (i >= start) {
      DEBUG(printf(" i=%d xy= %d,%d  (dxdy=%d,%d)\n", i, x,y, dx,dy));
      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == x && v_to_xy[to].y == y) {
          goto GOT_TO;
        }
      }
      DEBUG(printf("  new location\n"));
      to = xy_upto++;
      if (to >= n) {
        fprintf(stderr, "terdragon_graph() oops, at i=%d (end=%d) to=%d xy=%d,%d exceeds size n=%d\n",
                i, end, to, x,y, n);
        abort();
      }
      v_to_xy[to].x = x;
      v_to_xy[to].y = y;
    GOT_TO:
      if (from >= 0) {
        DEBUG(printf("  edge %d to %d at %d,%d\n", from,to, x,y));
        ADDONEEDGE(g, from, to, m);
      }
      from = to;
      terdragon_graph_INTERNAL_last_v = to;
    } else {
      DEBUG(printf(" skip i=%d xy= %d,%d\n", i, x,y));
    }

    x += dx;
    y += dy;
    if (terdragon_next_turn_is_right(i))
      TERDRAGON_XY_ROTATE_MINUS120(dx,dy);      /* right */
    else
      TERDRAGON_XY_ROTATE_PLUS120(dx,dy);       /* left */
  }

  if (xy_upto != n) {
    fprintf(stderr, "oops, terdragon graph start=%d end=%d xy_upto=%d but vertices n=%d\n",
            start,end, xy_upto, n);
    abort();
  }
  FREES(v_to_xy);
}
void
terdragon_graph(graph *g, int m, int n, int k)
{
  terdragon_graph_INTERNAL(g, m,n, 0, (int)pow(3,k));
}
void
terdragon_triple_graph(graph *g, int m, int n, int k)
{
  terdragon_graph_INTERNAL(g, m,n, (int)4*pow(3,k), (int)7*pow(3,k));
}


/* ------------------------------------------------------------------------- */
/* Terdragon Triangle Turn Tree */


/* Modify vpar by star replacement.
   1 -> 1,2,3
   2 -> 4,5,6
   n -> 3*n-2,..,3*n
   3*n+1 above root

       *
       |
       *
      / \
     1   *
     |
     2
    / \
   3   4

   */
int
star_replacement_vpar (int *vpar, int n)
{
  DEBUG(printf("star_replacement_vpar() ");
        nautyextra_print_vpar(vpar,n));
  DYNALLSTAT(int,arm,arm_sz);
  DYNALLOC1(int,arm,arm_sz,n+1,"try");
  memset(arm, '\0', (n+1)*sizeof(int));
  int above = 3*n+1;
  int v;
  for (v = n; v >= 1; v--) {
    int p = vpar[v];
    DEBUG(printf("  rep v=%d p=%d  to %d to %d\n", v, p, 3*v-2, 3*v));
    assert (arm[p] >= -2);
    if (!p) {
      DEBUG(printf("   above %d\n", above));
      vpar[above] = 0;
    }
    vpar[3*v-2] = (p ? 3*p + (arm[p]--) : above++);
    vpar[3*v-1] = 3*v-2;    /* 5*v-2 middle */
    vpar[3*v]   = 3*v-2;
  }
  return above-1;
}


/* ------------------------------------------------------------------------- */
/* Alternate Terdragon Graph */

/* Return true if the alternate terdragon curve turns right after segment n.
   The first segment is n=0.  */
int
alternate_terdragon_next_turn_is_right (int n)
{
  int ret = 0;
  while ((n%3)==2) {
    n = (n-2)/3;
    ret ^= 1;
  }
  return ret ^ (n%3);
}

/* alternate_terdragon_graph(g, m,n, k) sets graph g to alternate terdragon
   curve k.
   n should be the number of vertices alternate_terdragon_P(k).
   g should have space for size m,n.

   The current implementation follows the curve by its turn sequence and
   does a linear search for preceding x,y to notice a point revisited.  This
   is not very efficient but is easy and is fast enough since any subsequent
   real work on the graph takes much longer.
 */
int alternate_terdragon_graph_INTERNAL_last_v;
static void
alternate_terdragon_graph_INTERNAL(graph *g, int m,int n, int start,int end)
{
  struct { int x; int y; } *v_to_xy;
  int i;
  int from = -1;
  int dx = 2;
  int dy = 0;
  int xy_upto = 0;
  int x = 0;
  int y = 0;
  DEBUG(printf("alternate_terdragon_graph_INTERNAL n=%d  start=%d end=%d\n", n, start,end));

  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("alternate_terdragon_graph()");

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  for (i = 0; i <= end; i++) {
    int to;
    if (i >= start) {
      DEBUG(printf(" i=%d xy= %d,%d  (dxdy=%d,%d)\n", i, x,y, dx,dy));
      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == x && v_to_xy[to].y == y) {
          goto GOT_TO;
        }
      }
      DEBUG(printf("  new location\n"));
      to = xy_upto++;
      if (to >= n) {
        fprintf(stderr, "alternate_terdragon_graph() oops, at i=%d (end=%d) to=%d xy=%d,%d exceeds size n=%d\n",
                i, end, to, x,y, n);
        abort();
      }
      v_to_xy[to].x = x;
      v_to_xy[to].y = y;
    GOT_TO:
      if (from >= 0) {
        DEBUG(printf("  edge %d to %d at %d,%d\n", from,to, x,y));
        ADDONEEDGE(g, from, to, m);
      }
      from = to;
      alternate_terdragon_graph_INTERNAL_last_v = to;
    } else {
      DEBUG(printf(" skip i=%d xy= %d,%d\n", i, x,y));
    }

    x += dx;
    y += dy;
    if (alternate_terdragon_next_turn_is_right(i))
      TERDRAGON_XY_ROTATE_MINUS120(dx,dy);      /* right */
    else
      TERDRAGON_XY_ROTATE_PLUS120(dx,dy);       /* left */
  }

  if (xy_upto != n) {
    fprintf(stderr, "oops, alternate_terdragon graph start=%d end=%d xy_upto=%d but vertices n=%d\n",
            start,end, xy_upto, n);
    abort();
  }
  FREES(v_to_xy);
}
void
alternate_terdragon_graph(graph *g, int m, int n, int k)
{
  alternate_terdragon_graph_INTERNAL(g, m,n, 0, (int)pow(3,k));
}

/* Curve up to but not including the middle segment */
void
alternate_terdragon_half_graph(graph *g, int m, int n, int k)
{
  alternate_terdragon_graph_INTERNAL(g, m,n, 0, (int) (pow(3,k)-1)/2);
}



/* ------------------------------------------------------------------------- */

void
try_terdragon_triangle_turns_tree (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  DYNALLOC1(int,vpar,vpar_sz,2,"try");
  SG_DECL(sg);
  SG_INIT(sg);

  int n = 1;
  vpar[0] = -1;
  vpar[1] = 0;

  int k;
  for (k=2; k<=6; k++) {
    int new_n = 3*n+1;
    DYNREALLOC(int,vpar,vpar_sz,new_n+1,"try");
    n = star_replacement_vpar (vpar,n);
    assert (n==new_n);
    /* nautyextra_print_vpar(vpar,n); */
    /* nautyextra_view_vpar(vpar,n); */
    nautyextra_validate_vpar(vpar,n);
    nautyextra_vpar_to_sg(vpar,n, &sg, 0);  /* undirected */
    nautyextra_validate_sg(&sg);

    printf ("k=%d n=%d\n", k, n);

    /* indnum = TTDegCount not 3 (k) = 3^(k-1)
       count 1 unique
    */
    if (0) {
      int indnum = nautyextra_indnum_vpar(vpar,n);
      printf("  indnum %d\n", indnum);
    }
    if (0) {
      struct nautyextra_indnum_and_count  info;
      nautyextra_indnum_and_count_vpar(vpar,n, &info);
      printf("  indnum %d  count %Lu\n", info.indnum, info.count);
    }

    if (0) {
      /* indsets
         not in OEIS: 9,854,733427864
         factor(733427864)
      */
      unsigned long long indsets = nautyextra_indsets_count_vpar(vpar,n);
      printf("  indsets %Lu\n", indsets);
    }

    if (0) {
      /* domnum = TTDegCount3(k) = 1,4
         count                     1,4
         .                      k= 2,3
      */
      int domnum = nautyextra_domnum_vpar(vpar,n);
      printf("  domnum %d\n", domnum);
    }
    if (1) {
      /* k=2 n=4
         domnum 1  count 1
         k=3 n=13
         domnum 4  count 4
         k=4 n=40
         domnum 13  count 504
         k=5 n=121
         domnum 40  count 1008189504
         factor(504)
         factor(1008189504)
         504        == 2^3 * 3^2 * 7^1
         1008189504 == 2^6 * 3^8 * 7^4
      */
      struct nautyextra_domnum_and_count  info;
      /* nautyextra_domnum_and_count_sg(&sg, &info); */
      nautyextra_domnum_and_count_vpar(vpar,n, &info);
      printf("  domnum %d  count %Lu\n", info.domnum, info.count);
    }

    if (0) {
      /* matchnum = domnum = TTDegCount3(k) */
      int matchnum = nautyextra_matchnum_vpar(vpar,n);
      printf("  matchnum %d\n", matchnum);
    }
    if (0) {
      /* matchnum count 3,54,288684,43985701535789376,
         .              2  3      4                 5
         factor(1699990476912)
         GP-Test  1699990476912 == 2^4 * 13^4 * 3^12 * 7
         factor(288684)
         factor(43985701535789376)
      */
      struct nautyextra_matchnum_and_count info;
      nautyextra_matchnum_and_count_vpar(vpar,n, &info);
      printf("  matchnum %d count %Lu\n", info.matchnum, info.count);
    }

    if (0) {
      /* diameter = 2^k-2
         endpoints = 3*2^(k-2)
         vertices = 4,13,37,97,241
      */
      struct nautyextra_diameter_info info;
      nautyextra_diameter_info_vpar(vpar,n, &info);
      printf("  diameter %d endpoints %d vertices %d\n",
             info.diameter, info.num_endpoints, info.num_vertices);
    }
    if (0) {
      /* Wiener
         not in OEIS: 9, 270, 5913, 116154, 2178981
         terminal
         not in OEIS: 6,  78, 1086,  16746,  276594
      */
      struct nautyextra_Wiener_info info;
      struct nautyextra_Wiener_info terminal_info;
      nautyextra_Wiener_index_vpar(vpar,n, 0, &info);
      nautyextra_Wiener_index_vpar(vpar,n, 1, &terminal_info);
      printf("  W %Lu  terminal W %Lu\n",
             info.W, terminal_info.W);
    }
  }
}

void
writeautom (int *p, int n)
{
  int i;
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}

void
show_edges (graph *g, int m,int n)
{
  set *row = (set*) g;
  int from, to;
  printf("edges");
  for (from=0; from < n-1; from++, row+=m) {
    for (to=from+1; to < n; to++) {
      if (ISELEMENT(row,to))
        printf(" %d-%d", from,to);
    }
  }
  putchar('\n');
}

/* vector(10,n,turn(n)==-1)
 */
void
try_sequences (void)
{
  int n, k;
  printf("turn right n=1: ");
  for (n=0; n < 10; n++) {
    printf(" %d,", terdragon_next_turn_is_right(n));
  }
  printf("\n");

  for (k=0; k < 10; k++) {
    int P = terdragon_P(k);
    int TP = terdragon_TP(k);
    printf("k=%d  P=%d TP=%d\n", k, P, TP);
  }

  /* 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1 */
  printf("alternate turn right n=1: ");
  for (n=0; n < 16; n++) {
    printf(" %d,", alternate_terdragon_next_turn_is_right(n));
  }
  printf("\n");

}

/* Terdragon diameter
   1, 3, 5, 8, 13, 22, 39, 66, 113
   Terdragon Wiener
   1, 10, 61, 455, 4341, 46600, 572055, 7324521, 102274523


   Alternate Terdragon diameter
   not in OEIS: 1, 3, 3, 7, 9, 19, 27, 55, 81, 163
   radius
   not in OEIS: 1, 2, 2, 4, 5, 10, 14, 28, 41, 82
   Wiener
   not in OEIS: 1,10,56,465,4290,48547,572434,7671101,103351480,1515625351
   diameter odd 1 3 7 19 55 163
   A100702  1+2*3^(n-1)

   Alternate Terdragon half diameter
   not in OEIS: 1, 2, 3, 6, 10, 18, 31, 54, 94, 162
   radius  floor(sqrt(3)^k)
   1, 1, 2, 3, 5, 9, 16, 27, 47, 81
   Wiener
   not in OEIS: 1, 8, 66, 649, 7088, 86017, 1109899, 15310242, 3206057757



*/
void
try_terdragon_diameter (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;

  for (k=0; k <= 10; k++) {
    int radius = 0, diameter = 0, cycles = 0, diameter_formula = 0;
    long long W = 0;

    /* int n = terdragon_P(k); */
    /* int m = SETWORDSNEEDED(n); */
    /* DYNALLOC2(graph,g,g_sz,m,n,"malloc"); */
    /* terdragon_graph(g, m,n, k); */

    /* int n = terdragon_P(k); */
    /* int m = SETWORDSNEEDED(n); */
    /* DYNALLOC2(graph,g,g_sz,m,n,"malloc"); */
    /* alternate_terdragon_graph(g, m,n, k); */

    int n = terdragon_P(k);
    assert((n&1)==0);
    n >>= 1;
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    alternate_terdragon_half_graph (g, m,n, k);

    /* printf("k=%d  n=%d\n", k, n); */

    diamstats(g, m,n, &radius, &diameter);
    /* cycles = cyclecount(g, m,n); */
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(g, m,n, 0, &info);
    W = info.W;

    printf("k=%d n=%d  radius %d  diameter %d %d  cycles %d  W=%Ld\n",
           k, n, radius, diameter, diameter_formula, cycles, W);
  }
}

/* triple diameter
   1,2,4,7,13,22,40,67,121,202,

   Tdiam(k) = vector_modulo([3/2, 5/2],k) * 3^floor(k/2) - 1/2
   vector(12,k,k--; Tdiam(k))

   triple Wiener
   3,33,282,2856,33126,414618,5639850,78966105,1159377174,17124031896
 */
void
try_triple_diameter (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;

  for (k=0; k <= 7; k++) {
    int radius = 0, diameter = 0;
    int start_ecc = 0, start_ecc_destination = 0;
    int end_ecc = 0, end_ecc_destination = 0;
    long long W = 0;

    int n = terdragon_TP(k);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    terdragon_triple_graph(g, m,n, k);


    diamstats(g, m,n, &radius, &diameter);
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(g, m,n, 0, &info);
    W = info.W;
    start_ecc = nautyextra_eccentricity_of_vertex(g, m,n, 0, &start_ecc_destination);

    end_ecc = nautyextra_eccentricity_of_vertex(g, m,n, terdragon_graph_INTERNAL_last_v, &end_ecc_destination);

    printf("k=%d n=%d  radius %d  diameter %d  W=%Ld  start ecc %d (to %d)  end ecc %d (%d to %d)\n",
           k, n, radius, diameter, W,
           start_ecc, start_ecc_destination,
           end_ecc, terdragon_graph_INTERNAL_last_v, end_ecc_destination);
    /* if (n <= 8) { */
    /*   writeg6(stdout,g,m,n); */
    /*   writes6(stdout,g,m,n); */
    /*   show_edges(g,m,n); */
    /* } */

    {
      int v = 0, i;
      int *dist;
      dist = (int*) ALLOCS(n,sizeof(int));
      if (dist == NULL) alloc_error("for find_dist()");
      find_dist(g, m,n, v, dist);
      for (i = 0; i < n; i++) {
        if (dist[i] == start_ecc) {
          printf("  start_ecc attained by %d\n", i);
        }
      }
      FREES(dist);
    }
  }
}

/* ------------------------------------------------------------------------- */

void
test (void)
{
  int bad = 0;
  {
    int k;
    for (k = 0; k <= 6; k++) {
      DYNALLSTAT(graph, g, g_sz);
      int n = terdragon_P(k);
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g,g_sz,m,n,"malloc");
      terdragon_graph(g, m,n, k);

      {
        int by_nt = numtriangles(g,m,n);
        int by_A  = terdragon_A(k);
        if (! (by_nt == by_A)) {
          printf ("k=%d n=%d terdragon_A() = %d, triangles = %d\n",
                  k,n, by_A, by_nt);
          bad++;
        }
      }
    }
  }
  if (bad) abort();
}

int
main (void)
{
  setvbuf(stdout,NULL,_IONBF,0);

  test();
  try_terdragon_diameter();
  return 0;

  try_sequences();
  try_terdragon_triangle_turns_tree();
  try_triple_diameter();
  /* try_terdragon_area_tree_diameter(); */
  /* try_terdragon_area_tree(); */
  return 0;
}
