\\ Copyright 2016, 2017, 2018, 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Work in progress ... mostly runs, but going to change wildly.


\\ Usage: read("fracper.gp")
\\
\\ This is some code for representing fractions 0 <= f <= 1 as digits in a
\\ given base>=2.  A fraction has initial fixed digits, then repeating
\\ periodic digits.
\\
\\     [ df, df, df, ..., df, [ pf, pf, ..., pf ] ]
\\
\\ A fracper vector is the fixed digits followed by the last entry as the
\\ periodic digits.  If the fraction is exact in the given base then the
\\ periodic part is an empty vector.
\\
\\ Requires gp 2.9 up for use of Map().
\\
\\ Complex number fractions f partly work, maybe.  The base must be an
\\ integer.


\\-----------------------------------------------------------------------------
\\ Accessors

fracper_fixed(v) = v[1..#v-1];
{
addhelp(fracper_fixed,
"vec = fracper_fixed(fracper)
Return a vector of the fixed part of fracper.");
}

fracper_periodic(v) = v[#v];
{
addhelp(fracper_periodic,
"vec = fracper_periodic(fracper)
Return a vector of the periodic part of fracper.");
}


\\-----------------------------------------------------------------------------
\\ Helpers

\\ v is a vector.  Return a "rotated down" [ v[2], v[3], ..., v[#v], v[1] ].
\\ If v is less than 2 elements then return unchanged.
fracper_vector_rotate_down(v) = \
  if(#v<2,v, concat([v[2..-1],v[1]]));

\\ v is a vector.  Return a "rotated up" [ v[#v], v[1], v[2], ..., v[#v-1] ].
\\ If v is less than 2 elements then return unchanged.
fracper_vector_rotate_up(v,n=1) =
{
  if(#v<2 || !(n%=#v), v, n=#v-n; concat([v[n+1..#v], v[1..n]]));
}

\\ Like digits(n,base), but pad the result to be exactly k digits.
\\ If n is more than k digits then throw an error.
fracper_INTERNAL_digits_padded(n,k,base=10) =
{
  my(v = digits(n,base));
  if(#v>k, error("fracper_INTERNAL_digits_padded() n=",n,
                 " is bigger than k=",k," digits"));
  concat(vector(max(0,k-#v)),v);
}

\\-----------------------------------------------------------------------------
\\ v is a fracper vector [ digit, digit, ..., digit, [ digit, digit, ... ] ]
\\ which are digits interpreted as fractional
\\     f = 0.fffff (ppppp)repeat
\\ in binary, or the optional given "base".
\\
fracper_to_frac(fracper,base=10) =
{
  \\ print("fixed ",subst(Pol(v[1]),'x,base));
  \\ print("num ",subst(Pol(v[2]),'x,base));
  \\ print("den ",(base^length(v[2]) - 1));
  \\ print("den shift ",length(v[1]));
  \\ print("f ",subst(Pol(v[2]),'x,base) / (base^length(v[2]) - 1));

  my(p=fracper_periodic(fracper));
  fracper=fracper[^-1];  \\ fixed part
  my(f=subst(Pol(fracper),'x,base));
  if(#p,
     f += subst(Pol(p),'x,base) / (base^#p - 1) );
  f / base^#fracper; 
}
{
addhelp(fracper_to_frac,
"f = fracper_to_frac(fracper,base=10)
Return the fraction represented by fracper in the given base or decimal by
default.  base can be a complex number or quad etc for a non-integer base

base can be a polynomial variable to get a \"t_RFRAC\".  For example,
    fracper_to_frac([1,[3]],'x) == 1/x + 1/x*3/(x-1)
which is 1/x + 3/x^2 + 3/x^3 + 3/x^4 + ...");
}

\\ v is a vector.
\\ If it comprises a repeating [a,b,c, a,b,c, a,b,c] then return just [a,b,c].
\\ The shortest repeating initial part is taken.
\\ If v is all 0s then return [].
fracper_INTERNAL_vector_periodic_reduce(v) =
{
  if(#v,
     fordiv(#v,len,
           my(r=v[1..len]);
           forstep(i=len+1,#v,len,
                  if(r!=v[i..i+len-1], next(2)));
           return(if(r==[0],[],r))));
  v;
}

\\ v is a fracper vector [ d,d,...,d, [ d,d,...,d ] ] of fixed and periodic.
\\ Return a normalized v by removing any fixed parts which can be satisfied
\\ by a suitably rotated periodic.
\\ If the periodic part is empty then trailing fixed digit 0s are removed.
fracper_normalize(fracper) =
{
  my(p=fracper_INTERNAL_vector_periodic_reduce(fracper[#fracper]),
     pos=#p);
  forstep(i=#fracper-1,1,-1,
         if(fracper[i] != if(#p,p[pos],0),
            return(concat(fracper[1..i], [fracper_vector_rotate_up(p,pos)]))));
  [ fracper_vector_rotate_up(p,#fracper-1) ];
     
  \\ if(!#select(n->n!=0,fracper[#fracper]), fracper[#fracper]=[]);  \\ periodic empty if all zeros
  \\ while(#fracper>=2
  \\       && fracper[#fracper-1]==if(#fracper[#fracper], fracper[#fracper][#fracper[#fracper]], 0)
  \\       ,
  \\       fracper=fracper[^-2];
  \\       fracper[#fracper]=fracper_vector_rotate_up(fracper[#fracper]));
  \\ fracper;
}
{
addhelp(fracper_normalize,
"fracper = fracper_normalize(fracper)
Return fracper in a normalized form.
This means the fixed part is reduced if it can be expressed by the periodic,
the periodic part is reduced if it is something repeated, and if the
periodic is entirely zeros then empty.");
}

fracper_unroll(v,n=1) =
{
  \\ ENHANCE-ME: n divrem periodic length
  for(i=1,n,
     my(p=v[#v]);
     v=concat([ v[^-1], if(#p,p[1],0), [fracper_vector_rotate_down(p)]]));
  v;
}

\\ propagate carry ...
\\ fracper_add(v1,v2) =
\\ {
\\   v1=fracper_unroll(v1,max(0,#v2-#v1));
\\   v2=fracper_unroll(v2,max(0,#v1-#v2));
\\   if(v1[#v1]==[],v1[#v1]=[0]);
\\   if(v2[#v2]==[],v2[#v2]=[0]);
\\   my(len=lcm(#v1[#v1],#v2[#v2]));
\\   v1[#v1]=concat(vector(len/#v1[#v1],i,v1[#v1]));
\\   v2[#v2]=concat(vector(len/#v2[#v2],i,v2[#v2]));
\\   fracper_normalize(v1+v2);
\\ }


\\-----------------------------------------------------------------------------

\\ f is a rational 0 <= f <= 1.
\\ Return a fracper vector [ d,d,...,d, [ d,d,...,d ] ] of fixed and periodic parts.
\\ The last element is a vector of repeating digits, so
\\     f = 0.dddd (dddd)repeat
\\ If f is immediately periodic then the fixed part might be empty.
\\ If f has denominator is some base^n then the periodic part is empty or zeros.
\\
\\ There are two representations of 1/base (and similar denominators base^n),
\\ either the terminating form or an infinite [0, [base-1]].  Presently the
\\ terminating form is given, except for f=1 which is [ [base-1]] = 0.9999 etc.
\\
\\ frac_to_fracper(f, base=10) =
\\ {
\\   my(debug=0);
\\   if(debug,print("frac_to_fracper() ",f,"  base="base));
\\   if(!(real(f)>=0 && real(f)<=1 && imag(f)>=0 && imag(f)<=1),
\\      error("must have 0 <= f <= 1, got ",f));
\\   if(!(base>1), error("must have base > 1"));
\\   if(f==1,return([ [base-1]]));
\\ 
\\   my(num=numerator(f),
\\      den=denominator(f),
\\      g=gcd(base,den),
\\      dendiv=if(g>1,g,base),
\\      fixedlen=valuation(den,dendiv),
\\      fixed);
\\   num *= base^fixedlen;
\\   [fixed,num] = divrem(real(num),den) + I*divrem(imag(num),den);
\\   fixed=fracper_INTERNAL_digits_padded(fixed,fixedlen,base); \\ now fixed as digits
\\   if(debug,print(" gcd g="g));
\\   if(debug,print(" fixedlen "fixedlen" fixed "fixed" leaving num="num" den="den));
\\ 
\\   \\ if(num%dendiv==0,
\\   \\    num *= den;
\\   \\    my(more_fixed);
\\   \\    [more_fixed,num]=divrem(num,den);
\\   \\    more_fixed=fracper_INTERNAL_digits_padded(more_fixed,2,base);
\\   \\    if(debug,print(" more fixed "more_fixed" num="num));
\\   \\    fixed=concat(fixed,more_fixed));
\\ 
\\   concat(fixed,
\\          [if(num,
\\              my(len=znorder(Mod(base,den/dendiv^fixedlen)));
\\              if(debug,print(" len="len" power "base^len-1));
\\              fracper_INTERNAL_digits_padded((base^len-1)*num/den, len, base),
\\              [])]);
\\ }
           \\ return(concat(if(pos>1,digs[1..pos-1],[]),
           \\               [digs[pos..#digs]])));

frac_to_fracper(f, base=10) =
{
  my(debug=0);
  if(debug,print("frac_to_fracper f="f));
  if(!(real(f)>=0 && real(f)<=1 && imag(f)>=0 && imag(f)<=1),
     error("must have 0 <= f <= 1, got ",f));
  if(!(base>1), error("must have base > 1"));
  if(f==1,return([ [base-1]]));
  my(num = numerator(f),
     den = denominator(f),
     digit_list=List([]),
     seen   = Map(),
     seenat = List([]),
     d);
  if(debug,print("start num="num" den="den));

  while(num,
        num*=base;
        if(debug,print("at: num="num"  digit_list="digit_list" seen="seen));
        \\ print(divrem(real(num),den));
        \\ print(divrem(imag(num),den));
        [d,num] = divrem(real(num),den) + I*divrem(imag(num),den);
        if(debug,print("  division d="d" num="num));
        if(real(d)>=base, d-=1; num+=base);
        if(imag(d)>=base, d-=I; num+=I*base);
        listput(digit_list,d);
        if(mapisdefined(seen,num),
           my(pos=mapget(seen,num));
           digit_list=Vec(digit_list);
           if(debug,print("end: num="num"  pos="pos" digit_list="digit_list));
           return(fracper_normalize
              (concat(digit_list[1..pos], [digit_list[pos+1..-1]]))));
        mapput(seen,num,#digit_list););
  concat(Vec(digit_list), [[]]);
}
{
addhelp(frac_to_fracper,
"fracper = frac_to_fracper(f,base=10)
Return digits of fraction f in the given base as fracper periodic.
Must have 0 <= f < 1.
If f=n/base^k exact fraction in base then the fracper is its n digits and
empty periodic part.
Complex number f might work.  base must be rational > 1.");
}

\\-----------------------------------------------------------------------------

fracper_digit(fracper,n) = {
  if(n<#fracper, fracper[n],      \\ fixed part
     n-=#fracper;                 \\ index into periodic
     fracper=fracper[#fracper];   \\ periodic part
     if(#fracper, fracper[((n-1)%#fracper)+1], 0));
}
{
addhelp(fracper_digit,
"digit = fracper_digit(fracper,n)
Return the n'th digit of fracper.  The first digit is n=1, next n=2, etc.");
}

fracper_INTERNAL_compare(fracper1,fracper2,func) =
{
  my(p1=fracper_periodic(fracper1),
     p2=fracper_periodic(fracper2),
     pos1=2-#fracper1,
     pos2=2-#fracper2);
  for(i=1, max(#fracper1,#fracper2)-1 + lcm(max(1,#p1),max(#p2,1)),
     my(ret=func(if(pos1<=0, fracper1[#fracper1-1+pos1], #p1,p1[pos1], 0),
                 if(pos2<=0, fracper2[#fracper2-1+pos2], #p2,p2[pos2], 0)));
     if(ret,return(ret));
     if(pos1++>#p1, pos1=1);  \\ pre-increment, then wrap-around
     if(pos2++>#p2, pos2=1));
  0;
}

fracper_lex(fracper1,fracper2) = \
  fracper_INTERNAL_compare(fracper1,fracper2,lex);
{
addhelp(fracper_lex,
"order = fracper_lex(fracper1,fracper2)
Return -1, 0 or 1 for a lex() comparision of the digits of fracper1 and fracper2.
Digits are compared high to low.  If they differ anywhere then return the
lex() at that place.  If they are the same everywhere then return 0.

lex() can be used on any real numbers.");
}

fracper_cmp(fracper1,fracper2) = \
  fracper_INTERNAL_compare(fracper1,fracper2,cmp);
{
addhelp(fracper_cmp,
"order = fracper_cmp(fracper1,fracper2)
Return -1, 0 or 1 for a cmp() comparision of the digits of fracper1 and fracper2.
Digits are compared high to low.  If they differ anywhere then return the
cmp() at that place.  If they are the same everywhere then return 0.

cmp() is the \"universal\" compare.  It demands same type and value.  Use
simplify(fracper) if ");
}

fracper_is_equal(v1,v2) = (fracper_cmp(v1,v2)==0);
{
addhelp(fracper_is_equal,
"bool = fracper_is_equal(fracper1,fracper2)
Return 1 if fracper1==fracper2, meaning equal at every digit position.");
}

fracper_apply(func,fv) =
{
  for(i=1,#fv-1, fv[i]=func(fv[i]));
  my(pv=fv[#fv]);
  for(i=1,#pv, pv[i]=func(pv[i]));
  fv[#fv]=pv;
  fv;
}
{
addhelp(fracper_apply,
"fracper = fracper_apply(func,fracper)
Apply func to each digit of fracper and return the result as a new fracper.");
}

fracper_transform_statefunc(v,func,state=1) =
{
  for(i=1,#v-1,
      [state,v[i]]=func(state,v[i]));
  my(p=v[#v], state_seen=Map([state,1]), l);
  while(1,
        for(i=1,#p,
           my(out); [state,out]=func(state,p[i]);
           listput(l,out));
        my(pos);
        if(mapisdefined(state_seen,state,&pos),
           return(concat([v[1..#v-1], l[1..#pos-1],
                          [ l[pos..#l] ]
                         ]))));
}
{
addhelp(fracper_transform_statefunc,
"fracper = fracper_transform_statefunc(fracper,func,state=1)
Return fracper with its digits transformed according to func().

func is called [new_state,output_digit] = func(state, digit).
state is initially the given state parameter, then set to new_state after
each call.  digit is a digit from fracper.  output_digit is the
corresponding digit position for the output.

digits are transformed by func until a repeat of state and periodic position
is found.  A repeat state means the output digits repeat, thus determining
the periodic part of the output.

\"state\" values can be anything, but should eventually repeat on the
fracper periodic part.  state values are compared effectively with the
universal cmp() (due to a Map() used to watch for repeats).");
}

\\-----------------------------------------------------------------------------
