#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use warnings;
use FindBin;
use Math::BaseCnv 'cnv';
use Math::PlanePath::TerdragonCurve 118; # v.118 for xyxy_to_n_either()

BEGIN { unshift @INC, $FindBin::Bin }
use MyTerdragon '@dir6_to_dx','@dir6_to_dy','EpredL',
  'xy_rotate_minus120','xy_rotate_plus120','chamfer';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


print "  % generated by tools/$FindBin::Script\n";

my $path = Math::PlanePath::TerdragonCurve->new;

$Printwrap = 2;
my $n_lo = 373;
my $n_hi = 384;

my ($x_lo,$y_lo) = $path->n_to_xy($n_lo);
print "  \\coordinate (start) at ($x_lo,$y_lo);\n";

{
  # grey triangles

  $Printwrap = 2;
  printwrap("\\fill[my triangles]");

  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);

    my $rotate = (EpredL($n)
                  ? \&xy_rotate_plus120
                  : \&xy_rotate_minus120);

    my ($x2,$y2) = ($x+$dx, $y+$dy);
    my ($dx2,$dy2) = $rotate->($dx,$dy);

    my ($x3,$y3) = ($x2+$dx2, $y2+$dy2);
    my ($dx3,$dy3) = $rotate->($dx2,$dy2);

    my $cx = chamfer($dx,'\MyTriangle');
    my $cy = chamfer($dy,'\MyTriangle');
    my $cx2 = chamfer(-$dx,'\MyTriangle');
    my $cy2 = chamfer(-$dy,'\MyTriangle');
    my $cx3 = chamfer(-$dx2+$dx3,'\MyTriangleEnd');
    my $cy3 = chamfer(-$dy2+$dy3,'\MyTriangleEnd');

    ### dxdy: "$dx,$dy  $dx2,$dy2  $dx3,$dy3"
    ### cxcy: "$cx,$cy  $cx2,$cy2  $cx3,$cy3"

    printwrap_newline();
    $Printwrap = 4;
    printwrap("($x$cx,$y$cy)");
    $Printwrap = 6;
    printwrap("--($x2$cx2,$y2$cy2)");
    printwrap("--($x3$cx3,$y3$cy3)");
    printwrap("--cycle");
  }
  printwrap(";\n");
}

my $segments = MyLineSegments->new;
my $start_from;
{
  # segments which enclose, thick arrows

  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);
    my ($x2,$y2) = ($x+$dx, $y+$dy);
    $segments->add($x,$y, $x2,$y2);

    my $cx = chamfer($dx);
    my $cy = chamfer($dy);
    my $cx2 = chamfer(-$dx);
    my $cy2 = chamfer(-$dy);
    my $from = "$x$cx,$y$cy";
    my $to = "$x2$cx2,$y2$cy2";

    printwrap_newline();
    $Printwrap = 2;
    printwrap("\\draw[->,my segment]($from)");
    $Printwrap = 4;
    printwrap("--($to);");
    $start_from //= $from;
  }
  printwrap_newline();
}
{
  # chamfered connections between segments which enclose

  $Printwrap = 2;
  printwrap("\\draw[my segment join]");
  $Printwrap = 4;
  foreach my $n ($n_lo+1 .. $n_hi+1) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);
    my ($prev_dx,$prev_dy) = $path->n_to_dxdy($n-1);

    my $cx = chamfer($dx);
    my $cy = chamfer($dy);
    my $prev_cx = chamfer(-$prev_dx);
    my $prev_cy = chamfer(-$prev_dy);
    printwrap(" ($x$prev_cx,$y$prev_cy)");
    printwrap("--($x$cx,$y$cy)");

    if ($n == $n_hi+1) {
      printwrap("-- +($dx*\\MySegmentEnd,$dy*\\MySegmentEnd)");
    }
  }
  printwrap(";\n");
}

# within $n_hi
sub path_is_traversed {
  my ($x1,$y1, $x2,$y2) = @_;
  my $n = $path->xyxy_to_n_either ($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_hi;
}

{
  my @points;
  $Printwrap = 2;
  printwrap("\\draw[dashed]");
  $Printwrap = 4;
  foreach my $n ($n_lo .. $n_hi) {
    foreach my $dir6 (0 .. 5) {
      my ($x,$y) = $path->n_to_xy($n);
      my $dx = $dir6_to_dx[$dir6];
      my $dy = $dir6_to_dy[$dir6];
      my ($x2,$y2) = ($x+$dx, $y+$dy);

      next unless path_is_traversed($x,$y, $x2,$y2);
      next if $segments->is_traversed($x,$y, $x2,$y2);

      if ($x == $x_lo && $y == $y_lo) {
        my $cx = chamfer($dx);
        my $cy = chamfer($dy);
        my $from = "$x$cx,$y$cy";
        printwrap(" ($start_from)--($from)--($x2,$y2)");
      } else {
        printwrap(" ($x,$y) -- +($dx,$dy)");
      }
      $segments->add($x,$y, $x2,$y2);
      push @points, [$x2,$y2];
    }
  }

  foreach my $p (@points) {
    my ($x,$y) = @$p;
    foreach my $p2 (@points) {
      my ($x2,$y2) = @$p2;

      next unless path_is_traversed($x,$y, $x2,$y2);
      next if $segments->is_traversed($x,$y, $x2,$y2);
      printwrap(" ($x,$y)--($x2,$y2)");
      $segments->add($x,$y, $x2,$y2);
    }
  }

  foreach my $p (@points) {
    my ($x,$y) = @$p;
    foreach my $dir6 (0 .. 5) {
      my $dx = $dir6_to_dx[$dir6];
      my $dy = $dir6_to_dy[$dir6];
      my ($x2,$y2) = ($x+$dx, $y+$dy);

      next unless path_is_traversed($x,$y, $x2,$y2);
      next if $segments->is_traversed($x,$y, $x2,$y2);

      my $ex = chamfer($dx,'\MyEnd') || 0;
      my $ey = chamfer($dy,'\MyEnd') || 0;
      printwrap(" ($x,$y) -- +($ex,$ey)");
    }
  }

  printwrap(";\n");  # end of dashed
}

print "% GP-Test  Epred(",$n_lo-1,") == 0\n";
print "% GP-Test  Epred($n_lo) == 1\n";
print "% GP-Test  Epred($n_hi) == 1\n";
print "% GP-Test  Epred(",$n_hi+1,") == 0\n";
print "  % end generated\n";
exit 0;
