#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use FindBin;
use List::Util 'min';
use Math::PlanePath::TerdragonCurve 118; # v.118 for xyxy_to_n_either()

BEGIN { unshift @INC, $FindBin::Bin }
use MyTerdragon ':all';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;

my $max_level = $ARGV[0] || 7;
print "% generated by tools/$FindBin::Script $max_level\n";

my $path = Math::PlanePath::TerdragonCurve->new;

sub triangle_key {
  my ($x1,$y1, $x2,$y2, $x3,$y3) = @_;

  if ($y1 != $y2) {
    ($x1,$y1, $x2,$y2, $x3,$y3) = ($x2,$y2, $x3,$y3, $x1,$y1);
  }
  if ($x1 > $x2) {
    ($x1,$y1, $x2,$y2) = ($x2,$y2, $x1,$y1);
  }
  ### normalized triangle: "$x1,$y1 to $x2,$y2 to $x3,$y3"
  return "$x1,$y1";
}

my %type_to_factor = (GRT =>  6,    # 2 scale, 3 mean
                      GRSH => 4,    # 2 scale, 2 mean
                      GR   => 4,
                      GV   => 4,
                     );

my %values;
foreach my $type ('GV','GR','GRT','GRSH') {
  my $factor = $type_to_factor{$type} || die;

  printwrap_indent("% GP-DEFINE  ");
  printwrap("${type}_generated_samples = {[\n");
  $Printwrap = 2;
  foreach my $k (0 .. $max_level) {
    ### level: "k=$k  $type"
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my $count = 0;
    my $x_total = 0;
    my $y_total = 0;

    my $x_offset = 0;
    my $y_offset = 0;
    if ($type eq 'GV') {
      ($n_lo,$n_hi) = (10*$n_hi, 12*$n_hi);   # N=10,11,12 upward V
      ($x_offset,$y_offset) = $path->n_to_xy($n_hi);
    }

    ### $n_lo
    ### $n_hi
    ### $x_offset
    ### $y_offset

    my %seen;
    foreach my $n ($n_lo .. $n_hi-1) {
      next unless $path->_UNDOCUMENTED__n_segment_is_right_boundary($n);
      my ($x1,$y1) = $path->n_to_xy($n);
      $x1 -= $x_offset;
      $y1 -= $y_offset;

      my ($dx,$dy) = $path->n_to_dxdy($n);
      my $x2 = $x1 + $dx;
      my $y2 = $y1 + $dy;

      ($dx,$dy) = xy_rotate_minus120($dx,$dy);
      my $x3 = $x2 + $dx;
      my $y3 = $y2 + $dy;

      if ($type eq 'GR' || $type eq 'GV') {
        ### GRorV seg: "$x1 $y1 to $x2 $y2"
        $x_total += ($x1+$x2);
        $y_total += ($y1+$y2);

      } else {
        if ($seen{triangle_key($x1,$y1, $x2,$y2, $x3,$y3)}++) {
          ### duplicate ...
          next;
        }
        ### at: "n=$n  $x1,$y1    $x2,$y2    $x3,$y3"

        if ($type eq 'GRT') {
          $x_total += ($x1+$x2+$x3);
          $y_total += ($y1+$y2+$y3);

        } elsif ($type eq 'GRSH') {
          my $n = $path->xyxy_to_n_either($x2,$y2, $x3,$y3);
          if (defined $n && $n < $n_hi) {
            ### shortcut across to x3 ..
            $x_total += $x1 + $x3;
            $y_total += $y1 + $y3;
          } else {
            $x_total += $x1 + $x2;
            $y_total += $y1 + $y2;
          }
        }
      }
      $count++;
    }
    my $comma = ($k == $max_level ? '' : ',');
    printwrap("1/$factor/$count*($x_total + ($y_total)*sqrt3i)$comma\n");

    push @{$values{$type}->{'x'}}, $x_total;
    push @{$values{$type}->{'y'}}, $y_total;
    push @{$values{$type}->{'count'}}, $count;

    push @{$values{$type}->{'xw'}}, ($x_total - $y_total)/2;
  }
  $Printwrap = 0;
  printwrap("]};\n");
}
print "% end generated\n";


if (0) {
  print "\n";

  foreach my $type ('GRT','GRSH') {
    print "X = [ ", join(',', @{$values{$type}->{'x'}}), "]\n";
    print "Y = [ ", join(',', @{$values{$type}->{'y'}}), "]\n";
    print "XW = [ ", join(',', @{$values{$type}->{'xw'}}), "]\n";
  }

  foreach my $type ('GV','GR','GRT','GRSH') {
    print "% $type\n";
    my $factor = $type_to_factor{$type} || die;
    my $x_array = $values{$type}->{'x'};
    my $y_array = $values{$type}->{'y'};
    my $count_array = $values{$type}->{'count'};
    foreach my $i (0 .. min(4, $#$x_array)) {
      my $x = $x_array->[$i];
      my $y = $y_array->[$i];
      my $count = $count_array->[$i];
      my $denominator = $count * $factor;
      if ($y == 0) {
        printf "\\tfrac{%d}{%d}, \\;\n", $x, $denominator;
      } else {
        $y = sprintf "%+d", $y;
        if (abs($y) == 1) {
          $y =~ s/1//;
        }
        printf "\\tfrac{%d %s \\sqrt3 i}{%d}, \\;\n", $x, $y, $denominator;
      }
    }
  }
  foreach my $type ('GV','GR','GRT','GRSH') {
    print "\n% $type\n";
    my $factor = $type_to_factor{$type} || die;
    my $x_array = $values{$type}->{'x'};
    my $y_array = $values{$type}->{'y'};
    my $count_array = $values{$type}->{'count'};
    foreach my $i (0 .. min(4, $#$x_array)) {
      my $x = $x_array->[$i];
      my $y = $y_array->[$i];
      my $count = $count_array->[$i];
      $y = sprintf "%+d", $y;
      printf "%% GP-Test  %s(%d) == 1/%d/%d*(%d %s*sqrt3i)\n",
        $type,$i, $factor,$count, $x,$y;
    }
  }
}

exit 0;
