#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use FindBin;
use List::Util 'min';
use Lingua::EN::Numbers;
use Math::PlanePath::TerdragonCurve 118; # v.118 for xyxy_to_n_either()

BEGIN { unshift @INC, $FindBin::Bin }
use MyTerdragon ':all';
use MyPrintwrap ':all';


print "  % generated by tools/$FindBin::Script\n";
my $path = Math::PlanePath::TerdragonCurve->new;

sub transform {
  my ($x,$y) = @_;
  $y = "($y)*sqrt(3)";
  # ($x,$y) = ($y,"-($x)");  # rotate -90
  # ($x,$y) = ("-($y)",$x);  # rotate +90
  # ($x,$y) = ($y,$x);  # transpose
  return ("{$x},{$y}");
}

my $pending;
foreach my $level (0 .. 3) {
  print "  \\begin{scope}[shift=(pos-$level)]\n";
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);

  my $n_is_join = sub {
    my ($n) = @_;
    my ($x,$y) = $path->n_to_xy($n);
    my @n_list = $path->xy_to_n_list($x,$y);
    @n_list = grep {$_<=2*$n_hi} @n_list;
    my $below = grep {$_ < $n_hi} @n_list;
    my $above = grep {$_ > $n_hi} @n_list;
    if (! $above || ! $below) {
      return 0;
    }
    if ($n < $n_hi && $below == 2) {
      return 0;
    }
    if ($n > $n_hi && $above == 2) {
      return 0;
    }
    return 1;
  };

  my %seen_triangle;
  my $triangle = sub {
    my ($x1,$y1, $x2,$y2, $x3,$y3, $options) = @_;
    ($x1,$y1, $x2,$y2, $x3,$y3) = triangle_normalize($x1,$y1, $x2,$y2, $x3,$y3);

    #  x1,y1  ----  x2,y2                x3,y3
    #       \      /                    /      \
    #         x3,y3                x1,y1  ----  x2,y2

    return 0 if $seen_triangle{"$x1,$y1--$x2,$y2--$x3,$y3"}++;

    my ($yplus,$yminus);
    if ($y1 < $y3) {
      $yplus = '+';
      $yminus = '-';
    } else {
      $yplus = '-';
      $yminus = '+';
    }

    $options = (defined $options ? " [$options]" : "");
    $Printwrap = 4;
    $x1 = "$x1+3*\\MyMargin";
    $y1 = "$y1$yplus\\MyMargin";
    $x2 = "$x2-3*\\MyMargin";
    $y2 = "$y2$yplus\\MyMargin";
    $y3 = "$y3$yminus\\MyMargin*sqrt(3)";

    my $xy1 = transform($x1,$y1);
    my $xy2 = transform($x2,$y2);
    my $xy3 = transform($x3,$y3);

    printwrap("\\fill$options ($xy1)");
    $Printwrap += 2;
    printwrap("--($xy2)");
    printwrap("--($xy3)\n");
    printwrap("--cycle;\n");
    $Printwrap -= 2;
    return 1;
  };

  my $join = '';
  $Printwrap = 4;
  printwrap("\\draw ");
  $Printwrap += 2;
  foreach my $n ($n_lo .. 2*$n_hi) {
    my ($x,$y) = $path->n_to_xy($n);

    my $is_join = $n_is_join->($n);
    if ($is_join) {
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my ($prev_dx,$prev_dy) = $path->n_to_dxdy($n-1);
      my $sx = $x.chamfer(-$prev_dx);
      my $sy = $y.chamfer(-$prev_dy);
      my $tx = $x.chamfer($dx);
      my $ty = $y.chamfer($dy);
      my $sxy = transform($sx,$sy);
      my $txy = transform($tx,$ty);
      printwrap("$join($sxy)");
      printwrap("--($txy)");
    } else {
      my $xy = transform($x,$y);
      printwrap("$join($xy)");
    }
    $join = '--';
  }
  printwrap(";\n");

  $Printwrap = 4;
  foreach my $n ($n_lo, $n_hi, 2*$n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    my $xy = transform($x,$y);
    printwrap(" \\node at ($xy) [my dot];");
  }

  $Printwrap = 4;
  my $count = 0;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x1,$y1) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    my $dx = $x2-$x1;
    my $dy = $y2-$y1;

    foreach my $rot (\&xy_rotate_plus60,
                     \&xy_rotate_minus60) {
      my ($mdx,$mdy) = $rot->($dx,$dy);
      my $xm = $x1 + $mdx;
      my $ym = $y1 + $mdy;
      my $n2 = $path->xyxy_to_n_either($x1,$y1, $xm,$ym) // 0;
      my $n3 = $path->xyxy_to_n_either($x2,$y2, $xm,$ym) // 0;
      if ($n2 >= 2*$n_hi) { $n2 = 0; }
      if ($n3 >= 2*$n_hi) { $n3 = 0; }
      if ($n2 && $n3 && ($n2 >= $n_hi || $n3 >= $n_hi)) {
        my $n_part = ($n < $n_hi);
        my $n2_part = ($n2 < $n_hi);
        my $n3_part = ($n3 < $n_hi);
        if ($n_part != $n2_part
            || $n2_part != $n3_part
            || $n3_part != $n_part) {
          $count += $triangle->($x1,$y1, $x2,$y2, $xm,$ym, 'my join colour');
        }
      }
    }
  }
  printwrap_newline();

  my $name = Lingua::EN::Numbers::num2en($level);
  print "  \\end{scope}\n";
  print "  \\newcommand{\\MyJ$name}{$count}\n";
  $pending .= "% GP-Test-Last  J($level)==$count\n";
}
print $pending;
print "  % end generated\n";
exit 0;
